
/****************************************************************
 *
 *	rtp_asf.c		(C) Copyright Netgem 2008
 *	RTP ASF demuxer
 *
 ****************************************************************/

#include "avformat.h"
#include "rtp_internal.h"
#include "rtsp.h"
#include "libavutil/base64.h"
#include "asf.h"

typedef struct RTPAsfContext {
    AVFormatContext *input_ic; /* original input format context */
    AVFormatContext *asf_ic; /* asf format context */

    URLContext      *h;
    ByteIOContext   *pb;

    uint8_t         *header; /* asf header */
    int              header_len;
    AVPacket         pkt;
    int              pkt_position;
    int64_t         global_position;
} RTPAsfContext;

static int      rtp_asf_url_read(URLContext *h, uint8_t *buf, int buf_size);
static int64_t rtp_asf_url_seek(URLContext *h, int64_t offset, int whence);
static int      rtp_asf_url_close(URLContext *h);
static int      rtp_asf_url_read_pause(URLContext *h, int pause);
static int64_t rtp_asf_url_read_seek(URLContext *h, int stream_index,
                                      int64_t timestamp, int flags);
static int      rtp_asf_url_read_keepalive(URLContext *h);

static URLProtocol rtp_asf_protocol = {
    .name               = "rtp_asf",
    .url_read           = rtp_asf_url_read,
    .url_seek           = rtp_asf_url_seek,
    .url_close          = rtp_asf_url_close,
    .url_read_pause     = rtp_asf_url_read_pause,
    .url_read_seek      = rtp_asf_url_read_seek,
    .url_read_keepalive = rtp_asf_url_read_keepalive,
};

struct RTPAsfContext* rtp_asf_init(const char *header64)
{
    RTPAsfContext *ac;

    ac = av_mallocz(sizeof(*ac));
    if (!ac) {
        av_log(NULL, AV_LOG_ERROR, "cannot allocate rtp asf context\n");
        return NULL;
    }
    ac->header_len = strlen(header64);
    ac->header = av_malloc(ac->header_len);
    if (!ac->header) {
        av_log(NULL, AV_LOG_ERROR, "cannot allocate rtp asf header\n");
        goto fail;
    }
    ac->header_len = av_base64_decode(ac->header, header64, ac->header_len);
    if (ac->header_len < 0) {
        av_log(NULL, AV_LOG_ERROR, "invalid rtp asf header\n");
        goto fail;
    }
    return ac;

 fail:
    if (ac)
        av_free(ac->header);
    av_free(ac);
    return NULL;
}

int rtp_asf_open(RTPAsfContext *ac, AVFormatContext *input_ctx)
{
    int ret;
    AVInputFormat *inputFormat;
    AVFormatContext *asf_ctx = NULL;
    
    inputFormat = av_find_input_format("asf");
    if (!inputFormat) {
        av_log(input_ctx, AV_LOG_ERROR, "cannot find asf format\n");
        goto fail;
    }

    ac->h = av_mallocz(sizeof(*ac->h));
    if (!ac->h) {
        av_log(input_ctx, AV_LOG_ERROR, "cannot allocate asf protocol context\n");
        goto fail;
    }
    ac->h->prot = &rtp_asf_protocol;
    ac->h->priv_data = ac;
    ac->h->is_streamed = 1;

    ret = url_fdopen(&ac->pb, ac->h);
    if (ret < 0) {
        av_log(input_ctx, AV_LOG_ERROR, "cannot allocate byte IO context\n");
        goto fail;
    }

    ret = av_open_input_stream(&asf_ctx, ac->pb, input_ctx->filename,
                               inputFormat, NULL);
    if (ret < 0) {
        av_log(input_ctx, AV_LOG_ERROR, "cannot open asf stream\n");
        goto fail;
    }

    /* for now, input is used only inside rtp_asf */
    ac->input_ic = av_alloc_format_context();
    if (!ac->input_ic) {
        av_log(input_ctx, AV_LOG_ERROR, "cannot allocate input context\n");
        goto fail;
    }

    /* save rtsp context */
    *ac->input_ic = *input_ctx;
    /* replace original rtsp context with new asf context */
    *input_ctx = *asf_ctx;
    /* save asf context  */
    ac->asf_ic = input_ctx;
    /* remove surnumeral context */
    av_free(asf_ctx);

    return 0;

 fail:
    if (ac->asf_ic) {
        av_close_input_stream(ac->asf_ic);
        ac->asf_ic = NULL;
    }
    else {
        if (ac->pb)
            av_free(ac->pb->buffer);
        av_free(ac->pb);
        av_free(ac->h);
    }
    return -1;
}

void rtp_asf_close(RTPAsfContext *ac)
{
    if (ac->asf_ic)
        av_close_input_stream(ac->asf_ic);
    av_free_packet(&ac->pkt);
    av_free(ac->header);
    av_free(ac);
}

static int rtp_asf_url_read(URLContext *h, uint8_t *buf, int buf_size)
{
    RTPAsfContext *ac = h->priv_data;
    int len = -1;

    if (ac->global_position < ac->header_len) {
        len = FFMIN(buf_size, ac->header_len - ac->global_position);
        memcpy(buf, ac->header + ac->global_position, len);
        ac->global_position += len;
    }
    else if (ac->asf_ic) {

        if (ac->pkt_position >= ac->pkt.size) {
            int ret;

            av_free_packet(&ac->pkt);
            ac->pkt_position = 0;

            ret = av_read_frame(ac->input_ic, &ac->pkt);
            if (ret < 0) {
                av_log(ac->asf_ic, AV_LOG_ERROR, "Cannot read %s frame\n",
                       ac->input_ic->iformat->name);
                return ret;
            }
        }

        len = FFMIN(buf_size, ac->pkt.size - ac->pkt_position);
        memcpy(buf, ac->pkt.data + ac->pkt_position, len);
        ac->pkt_position += len;
        ac->global_position += len;
    }

    return len;
}

static int64_t rtp_asf_url_seek(URLContext *h, int64_t offset, int whence)
{
    RTPAsfContext *ac = h->priv_data;
    ASFContext *asf = ac->asf_ic->priv_data;

    av_log(ac->asf_ic, AV_LOG_DEBUG, "offset=%lld whence=%d\n", offset, whence);

    switch (whence) {
    case AVSEEK_SIZE:
        if (asf->hdr.flags & 0x01) {
            av_log(ac->asf_ic, AV_LOG_ERROR, "is_broadcast\n");
            return -1;
        }
        else if (asf->nb_packets)
            return ac->header_len + (asf->nb_packets * asf->packet_size);
        else
            return asf->hdr.file_size;

    case SEEK_SET:
        if (ac->global_position <= ac->header_len &&
            offset              <= ac->header_len) {
            /* seek inside header */
            ac->global_position = offset;
        }
        else {
            ac->global_position = offset;
            ac->pkt_position = ac->pkt.size;
        }
        break;
    case SEEK_CUR:
        if (ac->global_position          <= ac->header_len && 
            ac->global_position + offset >= 0              &&
            ac->global_position + offset <= ac->header_len) {
            /* seek inside header */
            ac->global_position += offset;
        }
        else if (ac->pkt_position + offset  >= 0 &&
                 ac->pkt_position + offset  <= ac->pkt.size) {
            /* seek inside current packet */
            ac->pkt_position += offset;
            ac->global_position += offset;
        }
        else {
            ac->global_position = offset;
            ac->pkt_position = ac->pkt.size;
        }
        break;

    case SEEK_END:
        if (asf->hdr.flags & 0x01) {
            av_log(ac->asf_ic, AV_LOG_ERROR, "is_broadcast\n");
            return -1;
        }
        else {
            av_log(ac->asf_ic, AV_LOG_DEBUG, "SEEK_END\n");
            /* TODO? */
        }
        break;

    default:
        av_log(ac->asf_ic, AV_LOG_ERROR,  "invalid whence\n");
        return -1;
    }

    return ac->global_position;
}

static int rtp_asf_url_close(URLContext *h)
{
    RTPAsfContext *ac = h->priv_data;

    if (ac->input_ic) {
        ac->asf_ic = NULL;
        av_close_input_stream(ac->input_ic);
    }
    else {
        av_free_packet(&ac->pkt);
        av_free(ac->header);
        av_free(ac);
    }

    return 0;
}

static int rtp_asf_url_read_pause(URLContext *h, int pause)
{
    RTPAsfContext *ac = h->priv_data;
    int a=0, b=0;

    /* update discard flag */
    while (a<ac->asf_ic->nb_streams && b<ac->input_ic->nb_streams) {
        AVStream *sta = ac->asf_ic->streams[a];
        AVStream *stb = ac->input_ic->streams[b];
        
        if (sta->codec->codec_type != CODEC_TYPE_VIDEO &&
            sta->codec->codec_type != CODEC_TYPE_AUDIO) {
            a++;
            continue;
        }
        if (stb->codec->codec_type != CODEC_TYPE_VIDEO &&
            stb->codec->codec_type != CODEC_TYPE_AUDIO) {
            b++;
            continue;
        }
       
        if (sta->codec->codec_type != stb->codec->codec_type)
            break; /* not in the same order ??? */
        
        stb->discard = sta->discard;
        a++;
        b++;
    }

    if (pause)
        return av_read_pause(ac->input_ic);
    else
        return av_read_play(ac->input_ic);
}

static int64_t rtp_asf_url_read_seek(URLContext *h, int stream_index,
                                      int64_t timestamp, int flags)
{
    RTPAsfContext *ac = h->priv_data;
    AVFormatContext *s = ac->input_ic;
    int i, ret;

    /* search for the first A/V input stream */
    for (i=0;i<s->nb_streams;i++) {
        if (s->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO ||
            s->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO)
            break;
    }
    if (i == s->nb_streams)
        return AVERROR(EINVAL);

    timestamp = av_rescale_q(timestamp,
                             ac->asf_ic->streams[stream_index]->time_base,
                             s->streams[i]->time_base);

    if (s->iformat->read_seek) {
        ret = s->iformat->read_seek(s, i, timestamp, flags);
        if (ret < 0)
            return ret;
        return ac->header_len;
    }
    else
        return AVERROR(ENOSYS);
}

static int rtp_asf_url_read_keepalive(URLContext *h)
{
    RTPAsfContext *ac = h->priv_data;
    return av_read_keepalive(ac->input_ic);
}

// return 0 on packet, no more left, 1 on packet, 1 on partial packet...
static int rtp_asf_handle_packet(PayloadContext *data,
                                 AVStream *st,
                                 AVPacket * pkt,
                                 uint32_t * timestamp,
                                 const uint8_t * buf,
                                 int len, int flags)
{
    int duration_field;
    int key_frame;
    
    if (len <= 6) {
        av_log(NULL, AV_LOG_ERROR, "invalid asf buffer length (%d)\n", len);
        return -1;
    }

    duration_field = buf[0] & 0x10;
    key_frame      = buf[0] & 0x80;

    buf += 4;
    len -= 4;

    if (duration_field) {
        buf += 2;
        len -= 2;
    }

    if (av_new_packet(pkt, len) < 0) {
        av_log(NULL, AV_LOG_ERROR, "cannot create asf packet with length %d\n", len);
        return -1;
    }

    memcpy(pkt->data, buf, len);
    return 0;
}

RTPDynamicProtocolHandler ff_asfvideo_dynamic_handler = {
    .enc_name             = "x-asf-pf",
    .codec_type           = CODEC_TYPE_VIDEO,
    .codec_id             = CODEC_ID_WMV3,
    .parse_packet         = rtp_asf_handle_packet,
};
RTPDynamicProtocolHandler ff_asfaudio_dynamic_handler = {
    .enc_name             = "x-asf-pf",
    .codec_type           = CODEC_TYPE_AUDIO,
    .codec_id             = CODEC_ID_WMAV2,
    .parse_packet         = rtp_asf_handle_packet,
};
