/*
 * arp.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/time.h>

#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <net/if.h>
#include <sys/time.h>
#include <netinet/ether.h>
#if !defined (USE_TINYLIBC)
#include <netpacket/packet.h>
#endif

#include "busybox.h"

#define APPLET_NAME "arping"

#define PROC_NET_ARP "/proc/net/arp"

static int check_column_caption(const char *column, char **buf)
{
    char *p;

    if (!buf || !*buf)
        return 0;
    p = *buf;
    while (isspace(*p))
        p++;
    while (*column) {
        if (*p++ != *column++)
            return 0;
    }
    if (*p && strchr(" \t\n", *p) == NULL)
        return 0;
    *buf = p;
    return 1;
}

static void print_arp_cache(void)
{
    FILE *f;
    char buf[256];
    char *p;

    f = fopen(PROC_NET_ARP, "r");
    if (!f) {
        perror(PROC_NET_ARP);
        return;
    }

    /* Check header */
    p = buf;
    if (!fgets(buf, sizeof(buf), f) ||
        !check_column_caption("IP address", &p) ||
        !check_column_caption("HW type", &p) ||
        !check_column_caption("Flags", &p) ||
        !check_column_caption("HW address", &p) ||
        !check_column_caption("Mask", &p) ||
        !check_column_caption("Device", &p)) {
        fprintf(stderr, "%s does not have the expected format\n",
                PROC_NET_ARP);
        fclose(f);
        return;
    }

    printf("Address\t\t\tHWtype\tHWaddress\t    Flags Mask\t\t  Iface\n");

    while (fgets(buf, sizeof(buf), f)) {
        char ipaddr[16], hwaddr[24], mask[16], iface[16];
        int hwtype, flags;
        char sflags[8], shwtype[8];

        if (sscanf(buf, "%16s %i %i %24s %16s %16s",
                   ipaddr, &hwtype, &flags, hwaddr, mask, iface) != 6)
            continue;

        if (hwtype == ARPHRD_ETHER)
            sprintf(shwtype, "ether");
        else
            sprintf(shwtype, "0x%5x", hwtype);

        sflags[0] = '\0';
        if (flags & ATF_COM)
            strcat(sflags, "C");
        if (flags & ATF_PERM)
            strcat(sflags, "M");
        if (flags & ATF_PUBL)
            strcat(sflags, "P");
        if (flags & ATF_DONTPUB)
            strcat(sflags, "!");
        if (flags & ATF_USETRAILERS)
            strcat(sflags, "T");

        if (!(flags & ATF_NETMASK))
            mask[0] = '\0';

        printf("%-23.23s\t", ipaddr);
        if (!(flags & ATF_COM)) {
            if (flags & ATF_PUBL)
                printf("%-8.8s%-20.20s", "*", "*");
            else
                printf("%-8.8s%-20.20s", "", "(incomplete)");
        } else {
            printf("%-8.8s%-20.20s", shwtype, hwaddr);
        }
        printf("%-6.6s%-15.15s %s\n", sflags, mask, iface);
    }

    fclose(f);
}

int arp_main(int argc, char **argv)
{
    char **p;
    int help = 0;

    for (p = argv + 1; *p; p++) {
        if (!strcmp(*p, "-h") ||
            !strcmp(*p, "-?")) {
            help = 1;
        help:
            printf("arp [-n]   Display arp cache\n"
                   "\n"
                   "arp -h     This help message\n"
                   "\n"
                   "  -n, --numeric  dont resolve names\n");
            return help ? 0 : 1;
        } else if (!strcmp(*p, "-n") ||
                   !strcmp(*p, "--numeric")) {
            /* This is what we do */
        } else if (!strcmp(*p, "-v") ||
                   !strcmp(*p, "--verbose")) {
            /* Means verbose output */
        } else if (!strcmp(*p, "-V")) {
            printf("netbox\n"
                   "arp 0.1\n");
            return 0;
        } else {
            fprintf(stderr, "unknown command/option %s\n", *p);
            goto help;
        }
    }

    print_arp_cache();
    return 0;
}
