/*
 * SDPL sdplextract.c: an utility to extract one or all partitions
 *                     from a sdpl layer
 *
 *   Copyright (C) 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * User mode extraction program
 *
 * This is a programm which extracts one or more partitions
 * from a given layer
 *
 * Usage: sdplextract [-a] [num] layer filename
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl */
/**SECTION: extraction tool */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef CONFIG_BLK_SDPL_64BITS
/* These are the GLIBC definition to use 64 bits file support */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <unistd.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/version.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <linux/sdpl.h>

/** Display usage
 */
static void usage (char *name)
{
        fprintf(stderr, "Usage: %s [-a] [num] layer filename\n", name);
}

/** Main entry point
 */
int main (int argc, char **argv)
{
    struct block_device bdev;
    sdpl_stat_t f_stat;
    Sector sect;
    /* main structure */
    struct SDPL_map *SDPL_map = NULL;
    /* File description */
    char *filename, *fdump = NULL, *layername;
    /* Flags */
#if !CONFIG_BLK_SDPL_EMBEDDED
    int ret = -1;
#endif
    int get_all = 0, part_min = 0, part_max = 0, numpart, dec;

    if (argc < 4) {
        usage(argv[0]);
        exit(1);
    }

    if (*argv[1] == '-' && *(argv[1] + 1) == 'a') {
        get_all = 1;
    } else {
        part_min = atoi(argv[1]);
        part_max = part_min + 1;
    }

    bdev.fd = -1;
    layername = argv[2];
    filename = argv[3];

    if (get_all == 1) {
        if ((fdump = malloc(strlen(filename) + 4)) == NULL) {
            fprintf(stderr, "Memory error\n");
            goto error;
        }
    } else {
        fdump = filename;
    }

    if ((sect.sector = malloc(512)) == NULL)
        goto error;
    if ((SDPL_map = malloc(sizeof(struct SDPL_map))) == NULL)
        goto error;
#if CONFIG_BLK_SDPL_EMBEDDED
    if ((bdev.fd = sdpl_open(layername, O_RDONLY | O_EXCL | O_SYNC)) < 0 ||
        sdpl_stat(layername, &f_stat) < 0) {
        goto error;
    }
#else /* CONFIG_BLK_SDPL_EMBEDDED */
    if ((bdev.fd = 
         sdpl_open(layername, O_RDWR | O_EXCL | O_SYNC | O_LARGEFILE)) < 0 ||
        sdpl_stat(layername, &f_stat) < 0) {
        goto error;
    }
    SDPL_map->infos = NULL;
    SDPL_map->pentries = NULL;
    SDPL_map->pmagic = NULL;
    SDPL_map->pflags = NULL;
    SDPL_map->psize = NULL;
    SDPL_map->poffset = NULL;
    SDPL_map->pnext = NULL;
    SDPL_map->pbloc = NULL;
    SDPL_map->bflags = NULL;
#endif /* CONFIG_BLK_SDPL_EMBEDDED */

    /* Get the MBR */
    if ((SDPL_map->mbr = 
         (SDPL_br_t *)read_dev_sector(&bdev, 0, &sect)) == NULL)
        goto error;
    SDPL_map->res_size = 1024;
    SDPL_map->device_size = f_stat.st_size;
    if (SDPL_map->device_size < 2048) {
        sdpl_error("Device too small: " STD_ATTRIB " bytes (2048)\n",
                   SDPL_map->device_size);
        goto error;
    }
    SDPL_map->st_size = f_stat.st_size - SDPL_map->res_size;
    if (get_SDPL_infos(SDPL_map) < 0) {
#if CONFIG_BLK_SDPL_EMBEDDED
        goto error;
#else /* CONFIG_BLK_SDPL_EMBEDDED */
        fix_SDPL_infos(SDPL_map);
        //        goto go_on;
#endif /* CONFIG_BLK_SDPL_EMBEDDED */
    }
    dump_SDPL_infos(SDPL_map);

    get_parts_infos(SDPL_map, sect, &bdev);
    /* There, we've got to extract all datas and dump them */
    if (get_all == 1) {
        part_min = 0;
        part_max = NPARTS(SDPL_map);
    }
    dec = SDPL_map->data_offset / 512;
    for (numpart = part_min; numpart < part_max; numpart++) {
        std_type bloc, sector, nsect, remap;
        struct block_device fdev;

        if (get_all == 1) {
            sprintf(fdump, "%s-%02d", filename, numpart);
        }
        sdpl_debug("Extract " STD_ATTRIB " blocs from part %d of %s (%d - "
                   "%08X - %08X) to file %s\n", SDPL_map->psize[numpart], 
                   numpart, 
                   layername, dec, SDPL_map->data_offset, 
                   SDPL_map->res_size, fdump);
        fdev.fd = sdpl_open(fdump, O_WRONLY | O_CREAT | O_EXCL);
        if (fchmod(fdev.fd, 0644) < 0) {
            fprintf(stderr, "can't set permissions for file %s\n", fdump);
            goto error;
        }

        for (bloc = 0; bloc < SDPL_map->psize[numpart]; bloc++) {
            sdpl_debug("Dump bloc " STD_ATTRIB " - " STD_ATTRIB "\n", 
                   bloc, SDPL_map->pbloc[numpart][bloc]);
            for (sector = 0; sector < SDPL_map->bloc_size; sector++) {
                nsect = SDPL_map->pbloc[numpart][bloc] * SDPL_map->bloc_size + 
                    sector + dec;
                remap = bloc * SDPL_map->bloc_size + sector;
                read_dev_sector(&bdev, nsect, &sect);
                write_dev_sector(&fdev, remap, &sect);
            }
        }
        close(fdev.fd);
    }

    ret = 0;
    goto end;

 error:
    if (sect.sector != NULL)
        sdpl_free(sect.sector);
    perror(argv[0]);

 end:

    if (bdev.fd > 0)
        close(bdev.fd);

    if (get_all == 1)
        free(fdump);

    return ret;
}
